from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_1 = StrOption(name="source_variable_1", doc="the first source variable", default="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_2 = StrOption(name="source_variable_2", doc="the second source variable", default="val2", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_3 = ChoiceOption(name="my_variable", doc="a variable", values=(Calculation(func['calc_value'], Params((ParamOption(option_1)))), Calculation(func['calc_value'], Params((ParamOption(option_2))))), default="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'choice'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1, option_2, option_3])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_4 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_3 = OptionDescription(name="type", doc="a type family", children=[option_4], properties=frozenset({"standard"}))
option_6 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_5 = OptionDescription(name="description", doc="This is a other great family", children=[option_6], properties=frozenset({"standard"}))
option_8 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_7 = OptionDescription(name="help", doc="a help family", children=[option_8], properties=frozenset({"standard"}), informations={'help': 'This is a other great family'})
option_10 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"advanced", "force_default_on_freeze", "frozen"}), informations={'type': 'string'})
optiondescription_9 = OptionDescription(name="mode", doc="a mode family", children=[option_10], properties=frozenset({"advanced"}))
option_12 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_11 = OptionDescription(name="hidden", doc="an hidden family", children=[option_12], properties=frozenset({"hidden", "standard"}))
option_14 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_13 = OptionDescription(name="disabled", doc="an disabled family", children=[option_14], properties=frozenset({"disabled", "standard"}))
optiondescription_2 = OptionDescription(name="my_family", doc="This is a great family", children=[optiondescription_3, optiondescription_5, optiondescription_7, optiondescription_9, optiondescription_11, optiondescription_13], properties=frozenset({"disabled", "hidden", "standard"}), informations={'help': 'This is a great family'})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_2], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_2 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
option_3 = StrOption(name="activer_ejabberd", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="général", children=[option_2, option_3], properties=frozenset({"normal"}))
optiondescription_18 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"normal"}))
option_5 = StrOption(name="description", doc="description", default="Exportation de la base de ejabberd", properties=frozenset({"mandatory", "normal"}))
option_6 = ChoiceOption(name="day", doc="day", values=('none', 'daily', 'weekly', 'monthly'), default=Calculation(func.calc_multi_condition, Params((ParamValue("non")), kwargs={'condition_1': ParamOption(option_3, notraisepropertyerror=True), 'match': ParamValue("none"), 'mismatch': ParamValue("daily")})), properties=frozenset({"mandatory", "normal"}))
option_7 = ChoiceOption(name="mode", doc="mode", values=('pre', 'post'), default="pre", properties=frozenset({"mandatory", "normal"}))
option_8 = StrOption(name="var1", doc="var1", properties=frozenset({"basic", "mandatory"}))
optiondescription_4 = OptionDescription(name="ejabberd", doc="ejabberd", children=[option_5, option_6, option_7, option_8], properties=frozenset({"basic"}))
optiondescription_19 = OptionDescription(name="extra", doc="extra", children=[optiondescription_4], properties=frozenset({"basic"}))
optiondescription_17 = OptionDescription(name="1", doc="1", children=[optiondescription_18, optiondescription_19])
option_10 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
option_11 = StrOption(name="activer_ejabberd", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_9 = OptionDescription(name="general", doc="général", children=[option_10, option_11], properties=frozenset({"normal"}))
optiondescription_21 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_9], properties=frozenset({"normal"}))
option_13 = StrOption(name="description", doc="description", default="Exportation de la base de ejabberd", properties=frozenset({"mandatory", "normal"}))
option_14 = ChoiceOption(name="day", doc="day", values=('none', 'daily', 'weekly', 'monthly'), default=Calculation(func.calc_multi_condition, Params((ParamValue("non")), kwargs={'condition_1': ParamOption(option_11, notraisepropertyerror=True), 'match': ParamValue("none"), 'mismatch': ParamValue("daily")})), properties=frozenset({"mandatory", "normal"}))
option_15 = ChoiceOption(name="mode", doc="mode", values=('pre', 'post'), default="pre", properties=frozenset({"mandatory", "normal"}))
option_16 = StrOption(name="var1", doc="var1", properties=frozenset({"basic", "mandatory"}))
optiondescription_12 = OptionDescription(name="ejabberd", doc="ejabberd", children=[option_13, option_14, option_15, option_16], properties=frozenset({"basic"}))
optiondescription_22 = OptionDescription(name="extra", doc="extra", children=[optiondescription_12], properties=frozenset({"basic"}))
optiondescription_20 = OptionDescription(name="2", doc="2", children=[optiondescription_21, optiondescription_22])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_17, optiondescription_20])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_4 = StrOption(name="var", doc="A dynamic variable", properties=frozenset({"basic", "mandatory", Calculation(func['variable_to_property'], Params((ParamValue("disabled"), ParamIdentifier()), kwargs={'when': ParamValue("val1"), 'inverse': ParamValue(False)}), help_function=func['variable_to_property'])}), informations={'type': 'string'})
optiondescription_3 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="dyn{{ identifier }}", identifiers=["val1", "val2"], children=[option_4], properties=frozenset({"basic"}))
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_3], properties=frozenset({"basic"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"basic"}))
option_8 = StrOption(name="var", doc="A dynamic variable", properties=frozenset({"basic", "mandatory", Calculation(func['variable_to_property'], Params((ParamValue("disabled"), ParamIdentifier()), kwargs={'when': ParamValue("val1"), 'inverse': ParamValue(False)}), help_function=func['variable_to_property'])}), informations={'type': 'string'})
optiondescription_7 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="dyn{{ identifier }}", identifiers=["val1", "val2"], children=[option_8], properties=frozenset({"basic"}))
optiondescription_6 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_7], properties=frozenset({"basic"}))
optiondescription_5 = OptionDescription(name="2", doc="2", children=[optiondescription_6], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_5])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries_old/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_3 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"expert", "mandatory"}))
optiondescription_2 = OptionDescription(name="general", doc="general", children=[option_3], properties=frozenset({"expert"}))
option_5 = ChoiceOption(name="enumvar", doc="multi", values=('c',), default="c", properties=frozenset({"expert", "mandatory"}))
option_5.impl_set_information('help', "bla bla bla")
optiondescription_4 = OptionDescription(name="enumfam", doc="enumfam", children=[option_5], properties=frozenset({"expert"}))
optiondescription_1 = OptionDescription(name="rougail", doc="rougail", children=[optiondescription_2, optiondescription_4], properties=frozenset({"expert"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



"""Takes a bunch of Rougail YAML dispatched in differents folders
as an input and outputs a Tiramisu's file.

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from itertools import chain
from pathlib import Path
from re import compile, findall
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    get_type_hints,
)

from pydantic import ValidationError
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from pydantic import ValidationError

from warnings import warn

from tiramisu.error import display_list

from .annotator import SpaceAnnotator
from .i18n import _
from .object_model import CONVERT_OPTION  # Choice,
from .object_model import (
    PROPERTY_ATTRIBUTE,
    CALCULATION_TYPES,
    CALCULATION_PROPERTY_TYPES,
    PARAM_TYPES,
    AnyParam,
    Calculation,
    Dynamic,
    Family,
    SymLink,
    Variable,
    VariableCalculation,
)
from .tiramisureflector import TiramisuReflector
from .utils import get_realpath, normalize_family, load_modules
from .error import DictConsistencyError

property_types = Union[Literal[True], Calculation]
properties_types = Dict[str, property_types]


class Property:
    def __init__(self) -> None:
        self._properties: Dict[str, properties_types] = {}

    def add(
        self,
        path: str,
        property_: str,
        value: property_types,
    ) -> None:
        self._properties.setdefault(path, {})[property_] = value

    def __getitem__(
        self,
        path: str,
    ) -> properties_types:
        return self._properties.get(path, {})

    def __contains__(
        self,
        path: str,
    ) -> bool:
        return path in self._properties


class Paths:
    regexp_relative = compile(r"^_*\.(.*)$")

    def __init__(
        self,
        default_namespace: str,
    ) -> None:
        self._data: Dict[str, Union[Variable, Family]] = {}
        self._dynamics: Dict[str:str] = {}
        if default_namespace is not None:
            default_namespace = normalize_family(default_namespace)
        self.default_namespace = default_namespace
        self.path_prefix = None

    def has_value(self) -> bool:
        return self._data != {}

    def add(
        self,
        path: str,
        data: Any,
        is_dynamic: bool,
        dynamic: str,
        *,
        force: bool = False,
    ) -> None:
        self._data[path] = data
        if not force and is_dynamic:
            self._dynamics[path] = dynamic

    def get_full_path(
        self,
        path: str,
        current_path: str,
    ):
        relative, subpath = path.split(".", 1)
        relative_len = len(relative)
        path_len = current_path.count(".")
        if path_len + 1 == relative_len:
            return subpath
        parent_path = current_path.rsplit(".", relative_len)[0]
        return parent_path + "." + subpath

    def get_with_dynamic(
        self,
        path: str,
        identifier_path: str,
        current_path: str,
        version: str,
        namespace: str,
        xmlfiles: List[str],
    ) -> Any:
        identifier = None
        if version != "1.0" and self.regexp_relative.search(path):
            path = self.get_full_path(
                path,
                current_path,
            )
        else:
            path = get_realpath(path, identifier_path)
        dynamic = None
        # version 1.0
        if version == "1.0":
            if not path in self._data and "{{ suffix }}" not in path:
                new_path = None
                current_path = None
                identifiers = []
                for name in path.split("."):
                    parent_path = current_path
                    if current_path:
                        current_path += "." + name
                    else:
                        current_path = name
                    if current_path in self._data:
                        if new_path:
                            new_path += "." + name
                        else:
                            new_path = name
                        continue
                    for dynamic_path in self._dynamics:
                        if "." in dynamic_path:
                            parent_dynamic, name_dynamic = dynamic_path.rsplit(".", 1)
                        else:
                            parent_dynamic = None
                            name_dynamic = dynamic_path
                        if (
                            parent_dynamic == parent_path
                            and name_dynamic.endswith("{{ identifier }}")
                            and name == name_dynamic.replace("{{ identifier }}", "")
                        ):
                            new_path += "." + name_dynamic
                            break
                        regexp = "^" + name_dynamic.replace("{{ identifier }}", "(.*)")
                        finded = findall(regexp, name)
                        if len(finded) != 1 or not finded[0]:
                            continue
                        if finded[0] == "{{ identifier }}":
                            identifiers.append(None)
                        else:
                            identifiers.append(finded[0])
                        if new_path is None:
                            new_path = name_dynamic
                        else:
                            new_path += "." + name_dynamic
                        parent_path = dynamic_path
                        break
                    else:
                        if new_path:
                            new_path += "." + name
                        else:
                            new_path = name
                path = new_path
            else:
                identifiers = None
                if "{{ suffix }}" in path:
                    path = path.replace("{{ suffix }}", "{{ identifier }}")
        elif not path in self._data:
            current_path = None
            parent_path = None
            new_path = current_path
            identifiers = []
            for name in path.split("."):
                if current_path:
                    current_path += "." + name
                else:
                    current_path = name
                # parent_path, name_path = path.rsplit('.', 1)
                if current_path in self._data:
                    if new_path:
                        new_path += "." + name
                    else:
                        new_path = name
                    parent_path = current_path
                    continue
                for dynamic_path in self._dynamics:
                    if "." in dynamic_path:
                        parent_dynamic, name_dynamic = dynamic_path.rsplit(".", 1)
                    else:
                        parent_dynamic = None
                        name_dynamic = dynamic_path
                    if (
                        "{{ identifier }}" not in name_dynamic
                        or parent_path != parent_dynamic
                    ):
                        continue
                    regexp = "^" + name_dynamic.replace("{{ identifier }}", "(.*)")
                    finded = findall(regexp, name)
                    if len(finded) != 1 or not finded[0]:
                        continue
                    if finded[0] == "{{ identifier }}":
                        identifiers.append(None)
                    else:
                        identifiers.append(finded[0])
                    if new_path is None:
                        new_path = name_dynamic
                    else:
                        new_path += "." + name_dynamic
                    parent_path = dynamic_path
                    break
                else:
                    if new_path:
                        new_path += "." + name
                    else:
                        new_path = name
                    if "{{ identifier }}" in name:
                        identifiers.append(None)
                    parent_path = current_path
            path = new_path
        else:
            identifiers = None
        if path not in self._data:
            return None, None
        option = self._data[path]
        option_namespace = option.namespace
        if (
            self.default_namespace not in [namespace, option_namespace]
            and namespace != option_namespace
        ):
            msg = _(
                'A variable or a family located in the "{0}" namespace shall not be used in the "{1}" namespace'
            ).format(option_namespace, namespace)
            raise DictConsistencyError(msg, 38, xmlfiles)
        return option, identifiers

    def __getitem__(
        self,
        path: str,
    ) -> Union[Family, Variable]:
        if not path in self._data:
            raise AttributeError(f"cannot find variable or family {path}")
        return self._data[path]

    def __contains__(
        self,
        path: str,
    ) -> bool:
        return path in self._data

    def __delitem__(
        self,
        path: str,
    ) -> None:
        logging.info("remove empty family %s", path)
        del self._data[path]

    def is_dynamic(self, path: str) -> bool:
        return path in self._dynamics

    def get(self):
        return self._data.values()


information_types = Dict[str, Union[str, int, float, bool]]


class Informations:
    def __init__(self) -> None:
        self._data: Dict[str, information_types] = {}

    def add(
        self,
        path: str,
        key: str,
        data: Any,
    ) -> None:
        if path not in self._data:
            self._data[path] = {}
        if key in self._data[path]:
            raise Exception(f'an information "{key}" is already present in "{path}"')
        self._data[path][key] = data

    def get(
        self,
        path: str,
    ) -> information_types:
        return self._data.get(path, {})


class ParserVariable:
    def __init__(self, rougailconfig):
        self.rougailconfig = rougailconfig
        self.load_config()
        self.paths = Paths(self.main_namespace)
        self.families = []
        self.variables = []
        self.parents = {".": []}
        self.index = 0
        self.reflector_names = {}
        self.leaders = []
        self.followers = []
        self.multis = {}
        self.default_multi = {}
        self.jinja = {}
        #
        self.convert_options = list(CONVERT_OPTION)
        self.convert_options.extend(self.custom_types)
        #
        self.exclude_imports = []
        self.informations = Informations()
        self.properties = Property()
        # self.choices = Appendable()
        self.has_dyn_option = False
        self.path_prefix = None
        self.is_init = False
        super().__init__()

    def load_config(self) -> None:
        rougailconfig = self.rougailconfig
        self.sort_dictionaries_all = rougailconfig["sort_dictionaries_all"]
        try:
            self.main_dictionaries = rougailconfig["main_dictionaries"]
        except:
            self.main_dictionaries = []
        self.main_namespace = rougailconfig["main_namespace"]
        if self.main_namespace:
            self.extra_dictionaries = rougailconfig["extra_dictionaries"]
        self.suffix = rougailconfig["suffix"]
        self.default_dictionary_format_version = rougailconfig[
            "default_dictionary_format_version"
        ]
        self.custom_types = rougailconfig["custom_types"]
        self.functions_files = rougailconfig["functions_files"]
        self.modes_level = rougailconfig["modes_level"]
        if self.modes_level:
            self.default_variable_mode = rougailconfig["default_variable_mode"]
            self.default_family_mode = rougailconfig["default_family_mode"]
        self.extra_annotators = rougailconfig["extra_annotators"]
        self.base_option_name = rougailconfig["base_option_name"]
        self.export_with_import = rougailconfig["export_with_import"]
        self.internal_functions = rougailconfig["internal_functions"]
        self.add_extra_options = rougailconfig[
            "structural_commandline.add_extra_options"
        ]
        self.plugins = rougailconfig["plugins"]

    def _init(self):
        if self.is_init:
            return
        variable = Variable
        family = Family
        if self.plugins:
            root = Path(__file__).parent
            for plugin in self.plugins:
                module_path = root / plugin / "object_model.py"
                if not module_path.is_file():
                    continue
                module = load_modules(
                    f"rougail.{plugin}.object_model", str(module_path)
                )
                if "Variable" in module.__all__:
                    variable = type(
                        variable.__name__ + "_" + plugin,
                        (variable, module.Variable),
                        {},
                    )
                if "Family" in module.__all__:
                    family = type(
                        family.__name__ + "_" + plugin, (family, module.Family), {}
                    )
        self.variable = variable
        self.family = family
        self.dynamic = type(Dynamic.__name__, (Dynamic, family), {})
        hint = get_type_hints(self.dynamic)
        # FIXME: only for format 1.0
        self.family_types = hint["type"].__args__  # pylint: disable=W0201
        self.family_attrs = frozenset(  # pylint: disable=W0201
            set(hint) - {"name", "path", "xmlfiles"} | {"redefine"}
        )
        self.family_calculations = self.search_calculation(  # pylint: disable=W0201
            hint
        )
        #
        hint = get_type_hints(self.variable)
        self.variable_types = (
            self.convert_options
        )  # hint["type"].__args__  # pylint: disable=W0201
        #
        self.variable_attrs = frozenset(  # pylint: disable=W0201
            set(hint) - {"name", "path", "xmlfiles"} | {"redefine", "exists"}
        )
        self.variable_calculations = self.search_calculation(  # pylint: disable=W0201
            hint
        )
        self.is_init = True

    ###############################################################################################
    # determine if the object is a family or a variable
    ###############################################################################################
    def is_family_or_variable(
        self,
        path: str,
        obj: dict,
        family_is_leadership: bool,
        version: str,
        filename: str,
    ) -> Literal["variable", "family"]:
        """Check object to determine if it's a variable or a family"""
        # it's already has a variable or a family
        if path in self.paths:
            if path in self.families:
                return "family"
            return "variable"
        # it's: "my_variable:"
        if not obj:
            return "variable"
        # check type attributes
        obj_type = self.get_family_or_variable_type(obj)
        if obj_type:
            if obj_type in self.family_types:
                return "family"
            if obj_type in self.variable_types:
                return "variable"
            msg = _("unknown type {0} for {1}").format(obj_type, path)
            raise DictConsistencyError(msg, 43, [filename])
        # in a leadership there is only variable
        if family_is_leadership:
            return "variable"
        # all attributes are in variable object
        # and values in attributes are not dict is not Calculation
        if isinstance(obj, dict):
            extra_keys = set(obj) - self.variable_attrs
            if not extra_keys:
                for key, value in obj.items():
                    if (
                        isinstance(value, dict)
                        and key != "params"
                        and not self.is_calculation(
                            key,
                            value,
                            self.variable_calculations,
                            False,
                        )
                    ):
                        break
                else:
                    return "variable"
        else:
            if version == "1.0":
                msg = f'Invalid value for the variable "{path}": "{obj}"'
                raise DictConsistencyError(msg, 102, [filename])
            return "variable"
        return "family"

    def get_family_or_variable_type(
        self,
        obj: dict,
    ) -> Optional[str]:
        """Check 'type' attributes"""
        if not isinstance(obj, dict):
            return None
        if "_type" in obj:
            # only family has _type attributs
            return obj["_type"]
        if "type" in obj and isinstance(obj["type"], str):
            return obj["type"]
        return None

    ###############################################################################################
    # create, update or delete family or variable object
    ###############################################################################################
    def family_or_variable(
        self,
        filename: str,
        name: str,
        subpath: str,
        obj: dict,
        version: str,
        comment: Optional[str],
        *,
        first_variable: bool = False,
        family_is_leadership: bool = False,
        family_is_dynamic: bool = False,
        parent_dynamic: Optional[str] = None,
    ) -> None:
        if name.startswith("_"):
            msg = f'the variable or family name "{name}" is incorrect, it must not starts with "_" character'
            raise DictConsistencyError(msg, 16, [filename])
        if not subpath:
            path = name
        else:
            path = f"{subpath}.{name}"
        if version == "0.1" and not isinstance(obj, dict) and obj is not None:
            msg = f'the variable "{path}" has a wrong type "{type(obj)}"'
            raise DictConsistencyError(msg, 17, [filename])
        typ = self.is_family_or_variable(
            path,
            obj,
            family_is_leadership,
            version,
            filename,
        )
        logging.info("family_or_variable: %s is a %s", path, typ)
        if typ == "family":
            parser = self.parse_family
        else:
            parser = self.parse_variable
        parser(
            filename,
            name,
            path,
            obj,
            version,
            comment=comment,
            first_variable=first_variable,
            family_is_leadership=family_is_leadership,
            family_is_dynamic=family_is_dynamic,
            parent_dynamic=parent_dynamic,
        )

    def parse_family(
        self,
        filename: str,
        name: str,
        path: str,
        obj: Optional[Dict[str, Any]],
        version: str,
        *,
        comment: Optional[str] = None,
        first_variable: bool = False,
        family_is_leadership: bool = False,
        family_is_dynamic: bool = False,
        parent_dynamic: Optional[str] = None,
    ) -> None:
        """Parse a family"""
        if obj is None:
            return
        family_obj = {}
        subfamily_obj = {}
        force_to_attrs = list(self.list_attributes(obj))
        for key, value in obj.items():
            if key in force_to_attrs:
                if key.startswith("_"):
                    key = key[1:]
                family_obj[key] = value
            else:
                subfamily_obj[key] = value
        if path in self.paths:
            # it's just for modify subfamily or subvariable, do not redefine
            if family_obj:
                if not obj.pop("redefine", False):
                    raise DictConsistencyError(
                        f'The family "{path}" already exists and it is not redefined',
                        32,
                        [filename],
                    )
                # convert to Calculation objects
                self.parse_parameters(
                    path,
                    obj,
                    filename,
                    family_is_dynamic,
                    False,
                    version,
                    typ="family",
                )
                self.paths.add(
                    path,
                    self.paths[path].model_copy(update=obj),
                    family_is_dynamic,
                    parent_dynamic,
                    force=True,
                )
            self.paths[path].xmlfiles.append(filename)
            force_not_first = True
            if self.paths[path].type == "dynamic":
                family_is_dynamic = True
                parent_dynamic = path
        else:
            if "redefine" in obj and obj["redefine"]:
                raise Exception(
                    f'cannot redefine the inexisting family "{path}" in {filename}'
                )
            extra_attrs = set(family_obj) - self.family_attrs
            if extra_attrs:
                raise Exception(f"extra attrs ... {extra_attrs}")
            obj_type = self.get_family_or_variable_type(family_obj)
            if obj_type is None:
                # auto set type
                if "_dynamic" in family_obj:
                    dynamic = family_obj["_dynamic"]
                elif "dynamic" in family_obj:
                    dynamic = family_obj["dynamic"]
                else:
                    dynamic = None
                if isinstance(dynamic, (list, dict)):
                    family_obj["type"] = obj_type = "dynamic"
            if obj_type == "dynamic":
                family_is_dynamic = True
                if "{{ identifier }}" not in name:
                    if version == "1.0" and "{{ suffix }}" in name:
                        name = name.replace("{{ suffix }}", "{{ identifier }}")
                        path = path.replace("{{ suffix }}", "{{ identifier }}")
                    elif "variable" in family_obj:
                        name += "{{ identifier }}"
                        path += "{{ identifier }}"
                    else:
                        msg = f'dynamic family name must have "{{{{ identifier }}}}" in his name for "{path}"'
                        raise DictConsistencyError(msg, 13, [filename])
                parent_dynamic = path
            if version != "1.0" and not family_obj and comment:
                family_obj["description"] = comment
            self.add_family(
                path,
                name,
                family_obj,
                filename,
                family_is_dynamic,
                parent_dynamic,
                version,
            )
            force_not_first = False
        if self.paths[path].type == "leadership":
            family_is_leadership = True
        for idx, key in enumerate(subfamily_obj):
            value = subfamily_obj[key]
            first_variable = not force_not_first and idx == 0
            comment = self.get_comment(key, obj)
            self.family_or_variable(
                filename,
                key,
                path,
                value,
                version,
                comment,
                first_variable=first_variable,
                family_is_leadership=family_is_leadership,
                family_is_dynamic=family_is_dynamic,
                parent_dynamic=parent_dynamic,
            )

    def list_attributes(
        self,
        obj: Dict[str, Any],
    ) -> Iterator[str]:
        """List attributes"""
        force_to_variable = []
        for key, value in obj.items():
            if key in force_to_variable:
                continue
            if key.startswith("_"):
                # if key starts with _, it's an attribute
                yield key
                # if same key without _ exists, it's a variable!
                true_key = key[1:]
                if true_key in obj:
                    force_to_variable.append(true_key)
                continue
            if isinstance(value, dict) and not self.is_calculation(
                key,
                value,
                self.family_calculations,
                False,
            ):
                # it's a dict, so a new variables!
                continue
            # 'variable' for compatibility to format 1.0
            if (
                key == "variable"
                and obj.get("type") != "dynamic"
                and obj.get("_type") != "dynamic"
            ):
                continue
            if key in self.family_attrs:
                yield key

    def add_family(
        self,
        path: str,
        name: str,
        family: dict,
        filename: str,
        family_is_dynamic: bool,
        parent_dynamic: str,
        version: str,
    ) -> None:
        """Add a new family"""
        family["path"] = path
        family["namespace"] = self.namespace
        family["version"] = version
        family["xmlfiles"] = [filename]
        obj_type = self.get_family_or_variable_type(family)
        if obj_type == "dynamic":
            family_obj = self.dynamic
            if version == "1.0":
                if "variable" not in family:
                    raise DictConsistencyError(
                        f'dynamic family must have "variable" attribute for "{path}"',
                        101,
                        family["xmlfiles"],
                    )
                if "dynamic" in family:
                    raise DictConsistencyError(
                        'variable and dynamic cannot be set together in the dynamic family "{path}"',
                        100,
                        family["xmlfiles"],
                    )
                family["dynamic"] = {
                    "type": "variable",
                    "variable": family["variable"],
                    "propertyerror": False,
                    "allow_none": True,
                }
                del family["variable"]
            # FIXME only for 1.0
            if "variable" in family:
                family["dynamic"] = {
                    "type": "variable",
                    "variable": family["variable"],
                    "propertyerror": False,
                    "allow_none": True,
                }
                del family["variable"]
                if version != "1.0":
                    warning = f'"variable" attribute in dynamic family "{ path }" is depreciated in {filename}'
                    warn(warning)
            if "variable" in family:
                raise Exception(
                    f'dynamic family must not have "variable" attribute for "{family["path"]}" in {family["xmlfiles"]}'
                )
        else:
            family_obj = self.family
        # convert to Calculation objects
        self.parse_parameters(
            path,
            family,
            filename,
            family_is_dynamic,
            False,
            version,
            typ="family",
        )
        try:
            self.paths.add(
                path,
                family_obj(name=name, **family),
                family_is_dynamic,
                parent_dynamic,
            )
        except ValidationError as err:
            raise Exception(f'invalid family "{path}" in "{filename}": {err}') from err
        self.set_name(
            self.paths[path],
            "optiondescription_",
        )
        if "." not in path:
            parent = "."
        else:
            parent = path.rsplit(".", 1)[0]
        self.parents[parent].append(path)
        self.parents[path] = []
        self.families.append(path)

    def parse_variable(
        self,
        filename: str,
        name: str,
        path: str,
        obj: Optional[Dict[str, Any]],
        version: str,
        *,
        comment: Optional[str] = None,
        first_variable: bool = False,
        family_is_leadership: bool = False,
        family_is_dynamic: bool = False,
        parent_dynamic: Optional[str] = None,
    ) -> None:
        """Parse variable"""
        if version == "1.0" or isinstance(obj, dict):
            if obj is None:
                obj = {}
            extra_attrs = set(obj) - self.variable_attrs
        else:
            extra_attrs = []
            obj = {"default": obj}
            if comment:
                obj["description"] = comment
        if extra_attrs:
            raise Exception(
                f'"{path}" is not a valid variable, there are additional '
                f'attributes: "{", ".join(extra_attrs)}"'
            )
        self.parse_parameters(
            path,
            obj,
            filename,
            family_is_dynamic,
            family_is_leadership is True and first_variable is False,
            version,
        )
        self.parse_params(path, obj)
        if path in self.paths:
            if "exists" in obj and not obj.pop("exists"):
                return
            if not obj.pop("redefine", False):
                msg = f'Variable "{path}" already exists'
                raise DictConsistencyError(msg, 45, [filename])
            self.paths.add(
                path,
                self.paths[path].model_copy(update=obj),
                family_is_dynamic,
                parent_dynamic,
                force=True,
            )
            self.paths[path].xmlfiles.append(filename)
        else:
            if "exists" in obj and obj.pop("exists"):
                # this variable must exist
                # but it's not the case
                # so do nothing
                return
            if "redefine" in obj and obj["redefine"]:
                msg = f'cannot redefine the inexisting variable "{path}"'
                raise DictConsistencyError(msg, 46, [filename])
            obj["path"] = path
            self.add_variable(
                name, obj, filename, family_is_dynamic, parent_dynamic, version
            )
            if family_is_leadership:
                if first_variable:
                    self.leaders.append(path)
                else:
                    self.followers.append(path)

    def parse_parameters(
        self,
        path: str,
        obj: dict,
        filename: str,
        family_is_dynamic: bool,
        is_follower: bool,
        version: str,
        *,
        typ: str = "variable",
    ):
        """Parse variable or family parameters"""
        if typ == "variable":
            calculations = self.variable_calculations
        else:
            calculations = self.family_calculations
        for key, value in obj.items():
            if self.is_calculation(
                key,
                value,
                calculations,
                False,
            ):
                try:
                    self.set_calculation(
                        obj,
                        key,
                        value,
                        path,
                        family_is_dynamic,
                        is_follower,
                        version,
                        [filename],
                    )
                except ValidationError as err:
                    raise Exception(
                        f'the {typ} "{path}" in "{filename}" has an invalid "{key}": {err}'
                    ) from err
                continue
            if not isinstance(value, list):
                continue
            for idx, val in enumerate(value):
                if not self.is_calculation(
                    key,
                    val,
                    calculations,
                    True,
                ):
                    continue
                try:
                    self.set_calculation(
                        obj,
                        key,
                        val,
                        path,
                        family_is_dynamic,
                        is_follower,
                        version,
                        [filename],
                        inside_list=True,
                        index=idx,
                    )
                except ValidationError as err:
                    raise Exception(
                        f'the {typ} "{path}" in "{filename}" has an invalid "{key}" '
                        f"at index {idx}: {err}"
                    ) from err

    def parse_params(self, path, obj):
        """Parse variable params"""
        if "params" not in obj:
            return
        if not isinstance(obj["params"], dict):
            raise Exception(f"params must be a dict for {path}")
        params = []
        for key, val in obj["params"].items():
            try:
                params.append(
                    AnyParam(
                        key=key,
                        value=val,
                        type="any",
                        path=None,
                        is_follower=None,
                        attribute=None,
                        family_is_dynamic=None,
                        xmlfiles=None,
                    )
                )
            except ValidationError as err:
                raise Exception(
                    f'"{key}" has an invalid "params" for {path}: {err}'
                ) from err
        obj["params"] = params

    def add_variable(
        self,
        name: str,
        variable: dict,
        filename: str,
        family_is_dynamic: bool,
        parent_dynamic: Optional[str],
        version: str,
    ) -> None:
        """Add a new variable"""
        if not isinstance(filename, list):
            filename = [filename]

        variable["namespace"] = self.namespace
        variable["version"] = version
        variable["path_prefix"] = self.path_prefix
        variable["xmlfiles"] = filename
        variable_type = self.get_family_or_variable_type(variable)
        obj = {
            "symlink": SymLink,
            "choice": self.variable,
        }.get(variable_type, self.variable)
        try:
            variable_obj = obj(name=name, **variable)
        except ValidationError as err:
            raise Exception(
                f'invalid variable "{variable["path"]}" in "{filename}": {err}'
            ) from err
        self.paths.add(
            variable["path"],
            variable_obj,
            family_is_dynamic,
            parent_dynamic,
        )
        self.variables.append(variable["path"])
        if "." in variable["path"]:
            parent_path = variable["path"].rsplit(".", 1)[0]
        else:
            parent_path = "."
        self.parents[parent_path].append(variable["path"])
        self.set_name(
            variable_obj,
            "option_",
        )

    def del_family(
        self,
        path: str,
    ) -> None:
        """The family is empty, so delete it"""
        del self.paths[path]
        self.families.remove(path)
        del self.parents[path]
        if "." in path:
            parent = path.rsplit(".", 1)[0]
        else:
            parent = "."
        self.parents[parent].remove(path)

    ###############################################################################################
    # set tiramisu file name
    ###############################################################################################
    def set_name(
        self,
        obj: Union[Variable, Family],
        option_prefix: str,
    ):
        """Set Tiramisu object name"""
        self.index += 1
        self.reflector_names[obj.path] = f"{option_prefix}{self.index}{self.suffix}"

    ###############################################################################################
    # calculations
    ###############################################################################################
    def is_calculation(
        self,
        attribute: str,
        value: dict,
        calculations: list,
        inside_list: bool,
    ):
        """Check if it's a calculation"""
        if inside_list:
            calculations = calculations[0]
        else:
            calculations = calculations[1]
        if not isinstance(value, dict) or attribute not in calculations:
            return False
        if "type" in value:
            return value["type"] in CALCULATION_TYPES
        # auto set type
        typ = set(CALCULATION_TYPES) & set(value)
        if len(typ) == 1:
            value["type"] = list(typ)[0]
            return True
        return False

    def set_calculation(
        self,
        obj: dict,
        attribute: str,
        value: dict,
        path: str,
        family_is_dynamic: bool,
        is_follower: bool,
        version: str,
        xmlfiles: List[str],
        *,
        inside_list: bool = False,
        index: int = None,
    ):
        """This variable is a calculation"""
        calculation_object = value.copy()
        typ = calculation_object.pop("type")

        calculation_object["attribute_name"] = attribute
        calculation_object["path_prefix"] = self.path_prefix
        calculation_object["path"] = path
        calculation_object["inside_list"] = inside_list
        calculation_object["version"] = version
        calculation_object["namespace"] = self.namespace
        calculation_object["xmlfiles"] = xmlfiles
        #
        if "params" in calculation_object:
            if not isinstance(calculation_object["params"], dict):
                raise Exception("params must be a dict")
            params = []
            for key, val in calculation_object["params"].items():
                if isinstance(val, dict) and "type" not in val:
                    # auto set type
                    param_typ = set(CALCULATION_TYPES) & set(val)
                    if len(param_typ) == 1:
                        val["type"] = list(param_typ)[0]
                if not isinstance(val, dict) or "type" not in val:
                    param_typ = "any"
                    val = {
                        "value": val,
                        "type": "any",
                    }
                else:
                    if version == "1.0" and val["type"] == "suffix":
                        val["type"] = "identifier"
                    param_typ = val["type"]
                val["key"] = key
                val["path"] = path
                val["family_is_dynamic"] = family_is_dynamic
                val["is_follower"] = is_follower
                val["attribute"] = attribute
                val["xmlfiles"] = xmlfiles
                try:
                    params.append(PARAM_TYPES[param_typ](**val))
                except ValidationError as err:
                    raise DictConsistencyError(
                        f'"{attribute}" has an invalid "{key}" for {path}: {err}',
                        29,
                        xmlfiles,
                    ) from err
            calculation_object["params"] = params
        #
        return_type = calculation_object.get("return_type")
        if return_type:
            if return_type not in self.variable_types:
                raise Exception(
                    f'unknown "return_type" in {attribute} of variable "{path}"'
                )
        #
        if typ == "identifier" and not family_is_dynamic:
            msg = f'identifier calculation for "{attribute}" in "{path}" cannot be set none dynamic family'
            raise DictConsistencyError(msg, 53, xmlfiles)
        if attribute in PROPERTY_ATTRIBUTE:
            calc = CALCULATION_PROPERTY_TYPES[typ](**calculation_object)
        else:
            calc = CALCULATION_TYPES[typ](**calculation_object)
        if index is None:
            obj[attribute] = calc
        else:
            obj[attribute][index] = calc


class RougailConvert(ParserVariable):
    """Main Rougail conversion"""

    supported_version = ["1.0", "1.1"]

    def __init__(self, rougailconfig) -> None:
        self.annotator = False
        self.yaml = YAML()
        super().__init__(rougailconfig)

    def search_calculation(
        self,
        hint: dict,
    ) -> Tuple[List[Any], List[Any]]:
        """attribute is calculated if typing is like: Union[Calculation, xxx]"""
        inside_list = []
        outside_list = []
        for key, value in hint.items():
            if "Union" in value.__class__.__name__ and (
                Calculation in value.__args__ or VariableCalculation in value.__args__
            ):
                outside_list.append(key)
            if (
                "Union" in value.__class__.__name__
                and "_GenericAlias" in value.__args__[0].__class__.__name__
                and Calculation in value.__args__[0].__args__
            ):
                inside_list.append(key)
            if (
                "Union" in value.__class__.__name__
                and value.__args__[0].__class__.__name__ == "_GenericAlias"
                and "Union" in value.__args__[0].__args__[0].__class__.__name__
                and Calculation in value.__args__[0].__args__[0].__args__
            ):
                inside_list.append(key)
        return inside_list, outside_list

    def parse_directories(
        self,
        path_prefix: Optional[str] = None,
    ) -> None:
        """Parse directories content"""
        self._init()
        if path_prefix:
            n_path_prefix = normalize_family(path_prefix)
            if n_path_prefix in self.parents:
                raise Exception("pfffff")
            root_parent = n_path_prefix
            self.path_prefix = n_path_prefix
            self.namespace = None
            self.add_family(
                n_path_prefix,
                n_path_prefix,
                {"description": path_prefix},
                "",
                False,
                None,
                "",
            )
        else:
            root_parent = "."
        if self.main_namespace:
            directory_dict = chain(
                (
                    (
                        self.main_namespace,
                        self.main_dictionaries,
                    ),
                ),
                self.extra_dictionaries.items(),
            )
            for namespace, extra_dirs in directory_dict:
                if namespace is None:
                    self.namespace = namespace
                else:
                    self.namespace = normalize_family(namespace)
                if root_parent == ".":
                    namespace_path = self.namespace
                else:
                    namespace_path = f"{root_parent}.{self.namespace}"
                if namespace_path in self.parents:
                    raise Exception("pfff")
                for idx, filename in enumerate(self.get_sorted_filename(extra_dirs)):
                    if not idx:
                        self.parse_family(
                            "",
                            self.namespace,
                            namespace_path,
                            {
                                "description": namespace,
                            },
                            "",
                        )
                    self.parse_variable_file(
                        filename,
                        namespace_path,
                    )
        else:
            self.namespace = None
            if root_parent == ".":
                namespace_path = ""
            else:
                namespace_path = f"{root_parent}"
            if namespace_path in self.parents:
                raise Exception("pfff")
            for filename in self.get_sorted_filename(self.main_dictionaries):
                self.parse_variable_file(
                    filename,
                    namespace_path,
                )
        if path_prefix:
            self.path_prefix = None

    def get_comment(
        self,
        name: str,
        objects: CommentedMap,
    ) -> Optional[str]:
        if name in objects.ca.items:
            comment = objects.ca.items[name][2]
        else:
            comment = None
        if comment:
            comment = comment.value[1:].strip()
        return comment

    def parse_variable_file(
        self,
        filename: str,
        path: str,
    ) -> None:
        """Parse file"""
        with open(filename, encoding="utf8") as file_fh:
            objects = self.yaml.load(file_fh)
        version = self.validate_file_version(
            objects,
            filename,
        )
        if objects is None:
            return
        self.parse_root_file(
            filename,
            path,
            version,
            objects,
        )

    def parse_root_file(
        self,
        filename: str,
        path: str,
        version: str,
        objects: dict,
    ) -> None:
        for name, obj in objects.items():
            comment = self.get_comment(name, objects)
            self.family_or_variable(
                filename,
                name,
                path,
                obj,
                version,
                comment,
            )

    def get_sorted_filename(
        self,
        directories: Union[str, List[str]],
    ) -> Iterator[str]:
        """Sort filename"""
        if not isinstance(directories, list):
            directories = [directories]
        if self.sort_dictionaries_all:
            filenames = {}
        for directory_name in directories:
            directory = Path(directory_name)
            if not directory.is_dir():
                continue
            if not self.sort_dictionaries_all:
                filenames = {}
            for file_path in directory.iterdir():
                if file_path.suffix not in [".yml", ".yaml"]:
                    continue
                if file_path.name in filenames:
                    raise DictConsistencyError(
                        _("duplicate dictionary file name {0}").format(file_path.name),
                        78,
                        [filenames[file_path.name][1]],
                    )
                filenames[file_path.name] = str(file_path)
            if not self.sort_dictionaries_all:
                for filename in sorted(filenames):
                    yield filenames[filename]
        if self.sort_dictionaries_all:
            for filename in sorted(filenames):
                yield filenames[filename]

    def validate_file_version(
        self,
        obj: dict,
        filename: str,
    ) -> None:
        """version is mandatory in YAML file"""
        if obj is None:
            obj = {}
        for name in ["_version", "version"]:
            if name not in obj:
                continue
            version = str(obj.pop(name))
            break
        else:
            # the `version` attribute is not mandatory
            default_version = self.default_dictionary_format_version
            if default_version is not None:
                version = default_version
            else:
                msg = '"version" attribut is mandatory in YAML file'
                raise DictConsistencyError(msg, 27, [filename])

        if version not in self.supported_version:
            msg = f'version "{version}" is not supported, list of supported versions: {display_list(self.supported_version, separator="or", add_quote=True)}'
            raise DictConsistencyError(msg, 28, [filename])
        return version

    def annotate(
        self,
    ):
        """Apply annotation"""
        if not self.paths.has_value():
            self.parse_directories()
        if self.annotator:
            raise DictConsistencyError(
                _("Cannot execute annotate multiple time"), 85, None
            )
        SpaceAnnotator(self)
        self.annotator = True

    def reflect(self) -> None:
        """Apply TiramisuReflector"""
        functions_files = [
            func for func in self.functions_files if func not in self.exclude_imports
        ]
        self.reflector = TiramisuReflector(
            self,
            functions_files,
        )

    def save(
        self,
        filename: str,
    ):
        """Return tiramisu object declaration as a string"""
        self._init()
        self.annotate()
        self.reflect()
        output = self.reflector.get_text() + "\n"
        if filename:
            with open(filename, "w", encoding="utf-8") as tiramisu:
                tiramisu.write(output)
        # print(output)
        return output



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['validators_int'] = "{% if _.int > 100 %}\nvalue is too high\n{% endif %}\n"
option_1 = IntOption(name="int", doc="A number", validators=[Calculation(func['valid_with_jinja'], Params((), kwargs={'__internal_jinja': ParamValue("validators_int"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_5validators/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("validators"), '__internal_variable': ParamValue("int"), '_.int': ParamSelfOption(whole=False)}))], properties=frozenset({"basic", "mandatory"}), informations={'type': 'number'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1])



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'Rougail'
copyright = '2019-2023, Silique'
author = 'gwen'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = '1.0'


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = [
        'sphinx.ext.extlinks', 'sphinx_lesson',
        #'myst_parser', 'sphinx.ext.extlinks'
]
#
#myst_enable_extensions = [
#    "amsmath",
#    "attrs_inline",
#    "colon_fence",
#    "deflist",
#    "dollarmath",
#    "fieldlist",
#    "html_admonition",
#    "html_image",
##    "linkify",
#    "replacements",
#    "smartquotes",
#    "strikethrough",
#    "substitution",
#    "tasklist",
#]


# **extlinks** 'sphinx.ext.extlinks',
# enables syntax like :proxy:`my source <hello>` in the src files
extlinks = {'proxy': ('/proxy/%s.html',
                      'external link: ')}

default_role = "code"

html_theme = "sphinx_rtd_theme"

pygments_style = 'sphinx'

html_short_title = "Rougail"
html_title = "Rougail documenation"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
#source_suffix = ['.rst', '.md']
source_suffix = '.rst'
#source_suffix = {
#    '.rst': 'restructuredtext',
#    '.txt': 'restructuredtext',
#    '.md': 'markdown',
#}
# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
# **themes**
#html_theme = 'bizstyle'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

def setup(app):
   app.add_css_file('css/custom.css')



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_1.rougail.dyn{{ identifier }}.var'] = "{% for val in __.var %}\nt{{ val }}\n{% endfor %}\n"
dict_env['default_1.rougail.dyn{{ identifier }}.dyn_{{ identifier }}.var_identifiers'] = "{{ s1 }}-{{ s2 }}\n"
dict_env['default_2.rougail.dyn{{ identifier }}.var'] = "{% for val in __.var %}\nt{{ val }}\n{% endfor %}\n"
dict_env['default_2.rougail.dyn{{ identifier }}.dyn_{{ identifier }}.var_identifiers'] = "{{ s1 }}-{{ s2 }}\n"
option_3 = StrOption(name="var", doc="A identifier variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_5 = StrOption(name="var", doc="A dynamic variable", multi=True, default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_1.rougail.dyn{{ identifier }}.var"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/60_6family_dynamic_sub_dynamic/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("1.rougail.dyn{{ identifier }}.var"), '__.var': ParamOption(option_3, notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_7 = StrOption(name="var", doc="A variable dynamic", default=Calculation(func['calc_value'], Params((ParamIdentifier()))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_8 = StrOption(name="var_identifier", doc="identifier from first family", default=Calculation(func['calc_value'], Params((ParamIdentifier(identifier_index=0)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_9 = StrOption(name="var_identifiers", doc="merge identifiers", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_1.rougail.dyn{{ identifier }}.dyn_{{ identifier }}.var_identifiers"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_6family_dynamic_sub_dynamic/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("1.rougail.dyn{{ identifier }}.dyn_{{ identifier }}.var_identifiers"), 's1': ParamIdentifier(identifier_index=0), 's2': ParamIdentifier(identifier_index=1)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_6 = ConvertDynOptionDescription(name="dyn_{{ identifier }}", doc="a Second dynamic variable", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_5)))), children=[option_7, option_8, option_9], properties=frozenset({"standard"}), informations={'dynamic_variable': '1.dyn{{ identifier }}.var'})
optiondescription_4 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="A dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_3)))), children=[option_5, optiondescription_6], properties=frozenset({"standard"}), informations={'dynamic_variable': '1.rougail.var'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, optiondescription_4], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_12 = StrOption(name="var", doc="A identifier variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_14 = StrOption(name="var", doc="A dynamic variable", multi=True, default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_2.rougail.dyn{{ identifier }}.var"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/60_6family_dynamic_sub_dynamic/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("2.rougail.dyn{{ identifier }}.var"), '__.var': ParamOption(option_12, notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_16 = StrOption(name="var", doc="A variable dynamic", default=Calculation(func['calc_value'], Params((ParamIdentifier()))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_17 = StrOption(name="var_identifier", doc="identifier from first family", default=Calculation(func['calc_value'], Params((ParamIdentifier(identifier_index=0)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_18 = StrOption(name="var_identifiers", doc="merge identifiers", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_2.rougail.dyn{{ identifier }}.dyn_{{ identifier }}.var_identifiers"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_6family_dynamic_sub_dynamic/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("2.rougail.dyn{{ identifier }}.dyn_{{ identifier }}.var_identifiers"), 's1': ParamIdentifier(identifier_index=0), 's2': ParamIdentifier(identifier_index=1)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_15 = ConvertDynOptionDescription(name="dyn_{{ identifier }}", doc="a Second dynamic variable", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_14)))), children=[option_16, option_17, option_18], properties=frozenset({"standard"}), informations={'dynamic_variable': '2.dyn{{ identifier }}.var'})
optiondescription_13 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="A dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_12)))), children=[option_14, optiondescription_15], properties=frozenset({"standard"}), informations={'dynamic_variable': '2.rougail.var'})
optiondescription_11 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_12, optiondescription_13], properties=frozenset({"standard"}))
optiondescription_10 = OptionDescription(name="2", doc="2", children=[optiondescription_11], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_10])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['validators_var1'] = "{% if values | length > 2 %}\nlength must be less than 3\n{% endif %}\n"
option_1 = StrOption(name="var1", doc="a second variable", multi=True, default=["no", "yes"], default_multi="no", validators=[Calculation(func['valid_with_jinja'], Params((), kwargs={'__internal_jinja': ParamValue("validators_var1"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_5validators_multi2/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("validators"), '__internal_variable': ParamValue("var1"), 'values': ParamSelfOption(whole=True)}))], properties=frozenset({"mandatory", "standard"}), informations={'type': 'string', 'test': ('val1', 'val2')})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['disabled_1.rougail.{{ identifier }}_dyn.var4'] = "{% if 'val4' not in __.var %}\nval4 is not a valid value\n{% endif %}\n"
dict_env['disabled_2.rougail.{{ identifier }}_dyn.var4'] = "{% if 'val4' not in __.var %}\nval4 is not a valid value\n{% endif %}\n"
option_3 = StrOption(name="var", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string', 'test': ('val1', 'val2', 'val3', 'val4')})
option_5 = StrOption(name="var1", doc="a variable 1", default=Calculation(func['calc_value'], Params((ParamIdentifier()))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_6 = StrOption(name="var2", doc="a variable 2", default=Calculation(func['calc_value'], Params((ParamOption(option_5)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_7 = StrOption(name="var3", doc="a variable 3", default=Calculation(func['calc_value'], Params((ParamOption(option_5)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_8 = StrOption(name="var4", doc="a variable 4", default=Calculation(func['calc_value'], Params((ParamDynOption(option_5, ["val4"])))), properties=frozenset({"mandatory", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("disabled")), kwargs={'__internal_jinja': ParamValue("disabled_1.rougail.{{ identifier }}_dyn.var4"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_unknown_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("disabled"), '__internal_variable': ParamValue("1.rougail.{{ identifier }}_dyn.var4"), 'when': ParamValue(True), 'inverse': ParamValue(False), '__.var': ParamOption(option_3, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_4 = ConvertDynOptionDescription(name="{{ identifier }}_dyn", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_3)))), children=[option_5, option_6, option_7, option_8], properties=frozenset({"standard"}), informations={'dynamic_variable': '1.rougail.var'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, optiondescription_4], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_11 = StrOption(name="var", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string', 'test': ('val1', 'val2', 'val3', 'val4')})
option_13 = StrOption(name="var1", doc="a variable 1", default=Calculation(func['calc_value'], Params((ParamIdentifier()))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_14 = StrOption(name="var2", doc="a variable 2", default=Calculation(func['calc_value'], Params((ParamOption(option_13)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_15 = StrOption(name="var3", doc="a variable 3", default=Calculation(func['calc_value'], Params((ParamOption(option_13)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_16 = StrOption(name="var4", doc="a variable 4", default=Calculation(func['calc_value'], Params((ParamDynOption(option_13, ["val4"])))), properties=frozenset({"mandatory", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("disabled")), kwargs={'__internal_jinja': ParamValue("disabled_2.rougail.{{ identifier }}_dyn.var4"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_unknown_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("disabled"), '__internal_variable': ParamValue("2.rougail.{{ identifier }}_dyn.var4"), 'when': ParamValue(True), 'inverse': ParamValue(False), '__.var': ParamOption(option_11, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_12 = ConvertDynOptionDescription(name="{{ identifier }}_dyn", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_11)))), children=[option_13, option_14, option_15, option_16], properties=frozenset({"standard"}), informations={'dynamic_variable': '2.rougail.var'})
optiondescription_10 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_11, optiondescription_12], properties=frozenset({"standard"}))
optiondescription_9 = OptionDescription(name="2", doc="2", children=[optiondescription_10], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_9])



"""Annotate dictionaries

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import importlib.resources
from os.path import isfile
from ..utils import load_modules


ANNOTATORS = None


def get_level(module):
    return module.level


def get_annotators(annotators, module_name, file_name=None):
    if file_name is None:
        _module_name = module_name
    else:
        _module_name = module_name + "." + file_name
        full_file_name = f"/{file_name}.py"
    annotators[_module_name] = []
    for pathobj in importlib.resources.files(module_name).iterdir():
        path = str(pathobj)
        if not path.endswith(".py") or path.endswith("__.py"):
            continue
        if file_name is not None and not path.endswith(full_file_name):
            continue
        module = load_modules(module_name, path)
        if "Annotator" not in dir(module):
            continue
        annotators[_module_name].append(module.Annotator)


class SpaceAnnotator:  # pylint: disable=R0903
    """Transformations applied on a object instance"""

    def __init__(
        self,
        objectspace,
    ):
        global ANNOTATORS
        if ANNOTATORS is None:
            ANNOTATORS = {}
            get_annotators(ANNOTATORS, "rougail.annotator")
        for extra_annotator in objectspace.extra_annotators:
            if extra_annotator in ANNOTATORS:
                continue
            get_annotators(ANNOTATORS, extra_annotator)
        for plugin in objectspace.plugins:
            try:
                get_annotators(ANNOTATORS, f"rougail.{plugin}", "annotator")
            except ModuleNotFoundError:
                pass
        annotators = ANNOTATORS["rougail.annotator"].copy()
        for extra_annotator in objectspace.extra_annotators:
            annotators.extend(ANNOTATORS[extra_annotator])
        for plugin in objectspace.plugins:
            annotators.extend(ANNOTATORS[f"rougail.{plugin}.annotator"])
        annotators = sorted(annotators, key=get_level)
        functions = {}
        functions_files = objectspace.functions_files
        for functions_file in functions_files:
            if isfile(functions_file):
                loaded_modules = load_modules("function_file", functions_file)
                for function in dir(loaded_modules):
                    if function.startswith("_"):
                        continue
                    functions[function] = getattr(loaded_modules, function)
        objectspace.functions = functions
        for annotator in annotators:
            annotator(objectspace)


__all__ = ("SpaceAnnotator",)



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_1.rougail.var1'] = "{% if var2 is defined %} {{ var2 }} {% elif var3 is defined %} {{ var3 }} {% elif var4 is defined %} {{ var4 }} {% else %} {{ _.var2 }} {% endif %} "
dict_env['default_2.rougail.var1'] = "{% if var2 is defined %} {{ var2 }} {% elif var3 is defined %} {{ var3 }} {% elif var4 is defined %} {{ var4 }} {% else %} {{ _.var2 }} {% endif %} "
option_4 = StrOption(name="var2", doc="a second variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_3 = StrOption(name="var1", doc="a first variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_1.rougail.var1"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/00_9default_calculation_param_optional/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("1.rougail.var1"), 'var2': ParamOption(option_4), '_.var2': ParamOption(option_4, notraisepropertyerror=True)})), properties=frozenset({"standard"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, option_4], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_8 = StrOption(name="var2", doc="a second variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_7 = StrOption(name="var1", doc="a first variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_2.rougail.var1"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/00_9default_calculation_param_optional/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("2.rougail.var1"), 'var2': ParamOption(option_8), '_.var2': ParamOption(option_8, notraisepropertyerror=True)})), properties=frozenset({"standard"}), informations={'type': 'string'})
optiondescription_6 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_7, option_8], properties=frozenset({"standard"}))
optiondescription_5 = OptionDescription(name="2", doc="2", children=[optiondescription_6], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_5])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_2 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
option_3 = StrOption(name="activer_ejabberd", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="général", children=[option_2, option_3], properties=frozenset({"normal"}))
optiondescription_38 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"normal"}))
option_6 = StrOption(name="description", doc="description", multi=True, default=['test'], properties=frozenset({"mandatory", "normal"}))
option_7 = ChoiceOption(name="mode", doc="mode", values=('pre', 'post'), multi=True, default_multi="pre", properties=frozenset({"mandatory", "normal"}))
optiondescription_5 = Leadership(name="description", doc="description", children=[option_6, option_7], properties=frozenset({"normal"}))
optiondescription_4 = OptionDescription(name="ejabberd", doc="ejabberd", children=[optiondescription_5], properties=frozenset({"normal"}))
optiondescription_39 = OptionDescription(name="extra", doc="extra", children=[optiondescription_4], properties=frozenset({"normal"}))
option_19 = FilenameOption(name="name", doc="name", default="/etc/mailname")
option_20 = StrOption(name="source", doc="source", default="mailname")
option_18 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_17 = OptionDescription(name="mailname", doc="mailname", children=[option_19, option_20, option_18])
option_23 = FilenameOption(name="name", doc="name", default="/etc/mailname2")
option_24 = StrOption(name="source", doc="source", default="mailname2")
option_22 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_21 = OptionDescription(name="mailname2", doc="mailname2", children=[option_23, option_24, option_22])
optiondescription_21.impl_set_information('engine', "jinja")
optiondescription_16 = OptionDescription(name="files", doc="files", children=[optiondescription_17, optiondescription_21])
option_15 = BoolOption(name="activate", doc="activate", default=True)
option_25 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_41 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_16, option_15, option_25])
optiondescription_41.impl_set_information('type', "service")
optiondescription_40 = OptionDescription(name="services", doc="services", children=[optiondescription_41], properties=frozenset({"hidden"}))
optiondescription_37 = OptionDescription(name="1", doc="1", children=[optiondescription_38, optiondescription_39, optiondescription_40])
option_9 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
option_10 = StrOption(name="activer_ejabberd", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_8 = OptionDescription(name="general", doc="général", children=[option_9, option_10], properties=frozenset({"normal"}))
optiondescription_43 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_8], properties=frozenset({"normal"}))
option_13 = StrOption(name="description", doc="description", multi=True, default=['test'], properties=frozenset({"mandatory", "normal"}))
option_14 = ChoiceOption(name="mode", doc="mode", values=('pre', 'post'), multi=True, default_multi="pre", properties=frozenset({"mandatory", "normal"}))
optiondescription_12 = Leadership(name="description", doc="description", children=[option_13, option_14], properties=frozenset({"normal"}))
optiondescription_11 = OptionDescription(name="ejabberd", doc="ejabberd", children=[optiondescription_12], properties=frozenset({"normal"}))
optiondescription_44 = OptionDescription(name="extra", doc="extra", children=[optiondescription_11], properties=frozenset({"normal"}))
option_30 = FilenameOption(name="name", doc="name", default="/etc/mailname")
option_31 = StrOption(name="source", doc="source", default="mailname")
option_29 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_28 = OptionDescription(name="mailname", doc="mailname", children=[option_30, option_31, option_29])
option_34 = FilenameOption(name="name", doc="name", default="/etc/mailname2")
option_35 = StrOption(name="source", doc="source", default="mailname2")
option_33 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_32 = OptionDescription(name="mailname2", doc="mailname2", children=[option_34, option_35, option_33])
optiondescription_32.impl_set_information('engine', "jinja")
optiondescription_27 = OptionDescription(name="files", doc="files", children=[optiondescription_28, optiondescription_32])
option_26 = BoolOption(name="activate", doc="activate", default=True)
option_36 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_46 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_27, option_26, option_36])
optiondescription_46.impl_set_information('type', "service")
optiondescription_45 = OptionDescription(name="services", doc="services", children=[optiondescription_46], properties=frozenset({"hidden"}))
optiondescription_42 = OptionDescription(name="2", doc="2", children=[optiondescription_43, optiondescription_44, optiondescription_45])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_37, optiondescription_42])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['disabled_rougail.variable1'] = "{% if _.condition == \"yes\" %}\ncondition is yes\n{% endif %}\n"
dict_env['disabled_rougail.variable2'] = "{% if _.condition == \"yes\" %}\ncondition is yes\n{% endif %}\n"
option_2 = StrOption(name="condition", doc="a conditional variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_3 = StrOption(name="variable1", doc="a first variable", properties=frozenset({"basic", "mandatory", Calculation(func['jinja_to_property'], Params((ParamValue("disabled")), kwargs={'__internal_jinja': ParamValue("disabled_rougail.variable1"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_5disabled_calculation/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("disabled"), '__internal_variable': ParamValue("rougail.variable1"), 'when': ParamValue(True), 'inverse': ParamValue(False), '_.condition': ParamOption(option_2, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
option_4 = StrOption(name="variable2", doc="a second variable", properties=frozenset({"basic", "mandatory", Calculation(func['jinja_to_property'], Params((ParamValue("disabled")), kwargs={'__internal_jinja': ParamValue("disabled_rougail.variable2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_5disabled_calculation/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("disabled"), '__internal_variable': ParamValue("rougail.variable2"), 'when': ParamValue(True), 'inverse': ParamValue(False), '_.condition': ParamOption(option_2, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, option_3, option_4], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from pytest import fixture, raises
from os import getcwd, listdir, environ, makedirs
from os.path import isfile, join, isdir, dirname
from shutil import rmtree, copyfile
import logging

environ['TIRAMISU_LOCALE'] = 'en'

from rougail import Rougail, RougailConfig
from rougail.error import DictConsistencyError
from .custom import CustomOption


logger = logging.getLogger()
logger.setLevel(logging.INFO)


dico_dirs = 'tests/dictionaries'

# if test_3_template.py failed, this temporary directory must be removed
tmp_dir = join(dico_dirs, 'tmp')
if isdir(tmp_dir):
    rmtree(tmp_dir)


test_ok = set()
test_raise = set()

for test in listdir(dico_dirs):
    if isdir(join(dico_dirs, test)):
        if isdir(join(dico_dirs, test, 'tiramisu')):
            test_ok.add(test)
        elif test != '__pycache__':
            test_raise.add(test)

excludes = set([
    '80family_several',
])
test_ok -= excludes
test_raise -= excludes
# test_ok = ['60_5family_dynamic_variable_outside_sub_suffix']
#test_ok = []
#test_raise = ['88valid_enum_invalid_default']
#test_raise = []
test_multi = True
#test_multi = False
test_base = test_multi
#test_base = True
#test_base = False
test_no_namespace = test_multi
#test_no_namespace = True
#test_no_namespace = False

ORI_DIR = getcwd()

debug = False
#debug = True

test_ok = list(test_ok)
test_raise = list(test_raise)
test_ok.sort()
test_raise.sort()

@fixture(scope="module", params=test_ok)
def test_dir(request):
    return request.param


@fixture(scope="module", params=test_raise)
def test_dir_error(request):
    return request.param


def get_tiramisu_filename(test_dir, subdir, multi, namespace):
    if not namespace and not multi:
        filename = 'no_namespace.py'
    elif not multi:
        filename = 'base.py'
    else:
        filename = 'multi.py'
    return join(test_dir, subdir, filename)


def load_rougail_object(test_dir, rougailconfig, multi=False, namespace=False):
    rougailconfig['functions_file'] = [join(dico_dirs, '../eosfunc/test.py')]
    dirs = [join(test_dir, 'dictionaries', 'rougail')]
    subfolder = join(test_dir, 'dictionaries', 'rougail2')
    if isdir(subfolder):
        dirs.append(subfolder)
    rougailconfig['dictionaries_dir'] = dirs
    extra_dictionaries = {}
    extras = listdir(join(test_dir, 'dictionaries'))
    extras.sort()
    for extra in extras:
        if extra in ['rougail', 'rougail2']:
            continue
        subfolder = join(test_dir, 'dictionaries', extra)
        if isdir(subfolder):
            extra_dictionaries[extra] = [subfolder]
    if extra_dictionaries:
        rougailconfig['extra_dictionaries'] = extra_dictionaries
    rougailconfig['tiramisu_cache'] = get_tiramisu_filename(test_dir, 'tmp', multi, namespace)
    rougailconfig['custom_types']['custom'] = CustomOption
    return Rougail(rougailconfig)


def save(test_dir, eolobj, multi=False, namespace=False, error=False):
    tiramisu_tmp = get_tiramisu_filename(test_dir, 'tmp', multi, namespace)
    tiramisu_tmp_dir = dirname(tiramisu_tmp)
    if isdir(tiramisu_tmp_dir):
        rmtree(tiramisu_tmp_dir)
    makedirs(tiramisu_tmp_dir)
    tiramisu_objects = eolobj.get_config()
    tiramisu_file = get_tiramisu_filename(test_dir, 'tiramisu', multi, namespace)
    tiramisu_dir = dirname(tiramisu_file)
    if not error:
        if not isdir(tiramisu_dir):
            raise Exception(f'creates {tiramisu_dir}')
        if not isfile(tiramisu_file) or debug:
            copyfile(tiramisu_tmp, tiramisu_file)
        with open(tiramisu_tmp, 'r') as fh:
            tiramisu_objects = fh.read()
        with open(tiramisu_file, 'r') as fh:
            tiramisu_objects_ori = fh.read()
        assert tiramisu_objects == tiramisu_objects_ori
        if isdir(tiramisu_tmp_dir):
            rmtree(tiramisu_tmp_dir)


def test_dictionary(test_dir):
    if not test_no_namespace:
        print('NAMESPACE!')
        return
    assert getcwd() == ORI_DIR
    test_dir_ = join(dico_dirs, test_dir)
    rougailconfig = RougailConfig.copy()
    rougailconfig['main_namespace'] = None
    if isfile(join(test_dir_, 'force_namespace')):
        return
    eolobj = load_rougail_object(test_dir_, rougailconfig)
    if not eolobj:
        return
    save(test_dir_, eolobj)
    assert getcwd() == ORI_DIR


def test_dictionary_namespace(test_dir):
    if not test_base:
        print('BASE!')
        return
    assert getcwd() == ORI_DIR
    test_dir_ = join(dico_dirs, test_dir)
    rougailconfig = RougailConfig.copy()
    rougailconfig['main_namespace'] = 'Rougail'
    eolobj = load_rougail_object(test_dir_, rougailconfig, namespace=True)
    if not eolobj:
        return
    save(test_dir_, eolobj, namespace=True)
    assert getcwd() == ORI_DIR


def test_dictionary_multi(test_dir):
    if not test_multi:
        print('MULTI!')
        return
    assert getcwd() == ORI_DIR
    test_dir_ = join(dico_dirs, test_dir)
    rougailconfig = RougailConfig.copy()
    rougailconfig['main_namespace'] = 'Rougail'
    eolobj = load_rougail_object(test_dir_, rougailconfig, multi=True)
    if not eolobj:
        return
    eolobj.add_path_prefix('1')
    eolobj.add_path_prefix('2')
    save(test_dir_, eolobj, multi=True)
    assert getcwd() == ORI_DIR


def test_error_dictionary(test_dir_error):
    assert getcwd() == ORI_DIR
    test_dir_ = join(dico_dirs, test_dir_error)
    errno = []
    rougailconfig = RougailConfig.copy()
    rougailconfig['main_namespace'] = 'Rougail'
    eolobj = load_rougail_object(test_dir_, rougailconfig, namespace=True)
    if eolobj is None:
        return
    for i in listdir(test_dir_):
        if i.startswith('errno_'):
            errno.append(int(i.split('_')[1]))
    if not errno:
        errno.append(0)
    with raises(DictConsistencyError) as err:
        save(test_dir_, eolobj, error=True)
    assert err.value.errno in errno, f'expected errno: {errno}, errno: {err.value.errno}, msg: {err}'
    tiramisu_tmp_dir = dirname(get_tiramisu_filename(test_dir_, 'tmp', False, True))
    if isdir(tiramisu_tmp_dir):
        rmtree(tiramisu_tmp_dir)
    assert getcwd() == ORI_DIR



from pytest import raises
import logging

from rougail import Rougail, RougailConfig
from rougail.error import DictConsistencyError

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def test_mode_invalid_default():
    # default variable mode is not in modes_level
    RougailConfig['dictionaries_dir'] = ['tests/personalize_mode/dictionary']
    RougailConfig['modes_level'] = ['level1', 'level2']
    with raises(ValueError) as err:
        RougailConfig['default_family_mode'] = 'level3'


def test_mode_invalid_default_family():
    # default family mode is not in modes_level
    RougailConfig['dictionaries_dir'] = ['tests/personalize_mode/dictionary']
    RougailConfig['modes_level'] = ['level1', 'level2']
    with raises(ValueError) as err:
        RougailConfig['default_variable_mode'] = 'level3'


def test_personalize_mode():
    RougailConfig['dictionaries_dir'] = ['tests/personalize_mode/dictionary']
    RougailConfig['modes_level'] = ['level1', 'level2']
    RougailConfig['default_variable_mode'] = 'level1'
    RougailConfig['default_family_mode'] = 'level1'
    RougailConfig['tiramisu_cache'] = None
    eolobj = Rougail()
    eolobj.get_config()


def test_personalize_mode_unknown():
    # a variable has an unknown mode
    RougailConfig['dictionaries_dir'] = ['tests/personalize_mode/dictionary']
    RougailConfig['modes_level'] = ['level1']
    RougailConfig['default_variable_mode'] = 'level1'
    RougailConfig['default_family_mode'] = 'level1'
    eolobj = Rougail()
    with raises(DictConsistencyError) as err:
        eolobj.converted.annotate()
    assert err.value.errno == 71


def test_personalize_annotate_twice():
    RougailConfig['dictionaries_dir'] = ['tests/personalize_mode/dictionary']
    RougailConfig['modes_level'] = ['level1', 'level2']
    RougailConfig['default_variable_mode'] = 'level1'
    RougailConfig['default_family_mode'] = 'level1'
    eolobj = Rougail()
    eolobj.converted.annotate()
    with raises(DictConsistencyError) as err:
        eolobj.converted.annotate()
    assert err.value.errno == 85



"""Annotate family

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Optional
from rougail.i18n import _
from rougail.error import DictConsistencyError
from rougail.annotator.variable import Walk
from rougail.object_model import VariableCalculation


class Mode:  # pylint: disable=R0903
    """Class to manage mode level"""

    def __init__(
        self,
        level: int,
    ) -> None:
        self.level = level

    def __gt__(
        self,
        other: int,
    ) -> bool:
        return other.level < self.level


class Annotator(Walk):
    """Annotate family"""

    level = 80

    def __init__(
        self,
        objectspace,
        *args,
    ):
        self.mode_auto = []
        self.objectspace = objectspace
        if not self.objectspace.paths:
            return
        self.check_leadership()
        self.remove_empty_families()
        self.family_description()
        if self.objectspace.modes_level:
            self.modes = {
                name: Mode(idx) for idx, name in enumerate(self.objectspace.modes_level)
            }
            self.default_variable_mode = self.objectspace.default_variable_mode
            self.default_family_mode = self.objectspace.default_family_mode
            self.change_modes()
        else:
            for family in self.get_families():
                self.valid_mode(family)
            for variable in self.get_variables():
                self.valid_mode(variable)
        self.convert_help()

    def check_leadership(self) -> None:
        """No subfamily in a leadership"""
        for family in self.get_families():
            if family.type != "leadership":
                continue
            for variable_path in self.objectspace.parents[family.path]:
                variable = self.objectspace.paths[variable_path]
                if variable.type in self.objectspace.family_types:
                    msg = f'the leadership "{family.path}" cannot have the { variable.type } "{ variable.path}"'
                    raise DictConsistencyError(msg, 24, variable.xmlfiles)

    def remove_empty_families(self) -> None:
        """Remove all families without any variable"""
        removed_families = []
        for family in self.get_families():
            if isinstance(family, self.objectspace.family) and not self._has_variable(
                family.path
            ):
                if (
                    self.objectspace.paths.default_namespace is None
                    or "." in family.path
                ):
                    removed_families.append(family.path)
        removed_families.reverse()
        for family in removed_families:
            self.objectspace.del_family(family)

    def _has_variable(
        self,
        family: str,
    ) -> bool:
        for variable in self.objectspace.parents[family]:
            if variable in self.objectspace.families:
                if self._has_variable(variable):
                    return True
            else:
                return True
        return False

    def family_description(self) -> None:
        for family in self.get_families():
            if not family.description:
                family.description = family.name
            if family.type == "dynamic" and isinstance(
                family.dynamic, VariableCalculation
            ):
                path = self.objectspace.paths.get_full_path(
                    family.dynamic.variable,
                    family.path,
                )
                self.objectspace.informations.add(family.path, "dynamic_variable", path)

    def change_modes(self):
        """change the mode of variables"""
        modes_level = self.objectspace.modes_level
        default_variable_mode = self.default_variable_mode
        if default_variable_mode not in modes_level:
            msg = _(
                'default variable mode "{0}" is not a valid mode, valid modes are {1}'
            ).format(default_variable_mode, modes_level)
            raise DictConsistencyError(msg, 72, None)
        default_family_mode = self.default_family_mode
        if default_family_mode not in modes_level:
            msg = _(
                'default family mode "{0}" is not a valid mode, valid modes are {1}'
            ).format(default_family_mode, modes_level)
            raise DictConsistencyError(msg, 73, None)
        families = list(self.get_families())
        for family in families:
            self.valid_mode(family)
            self._set_default_mode(family)
        families.reverse()
        for family in families:
            self._change_family_mode(family)
        if self.objectspace.paths.default_namespace is None:
            for variable_path in self.objectspace.parents["."]:
                variable = self.objectspace.paths[variable_path]
                if (
                    variable.type == "symlink"
                    or variable_path in self.objectspace.families
                ):
                    continue
                self._set_default_mode_variable(
                    variable,
                    self.default_variable_mode,
                    check_level=False,
                )

    def valid_mode(
        self,
        obj,
    ) -> None:
        modes_level = self.objectspace.modes_level
        if self._has_mode(obj) and obj.mode not in modes_level:
            if modes_level:
                msg = _(
                    'mode "{0}" for "{1}" is not a valid mode, valid modes are {2}'
                ).format(obj.mode, obj.name, modes_level)
            else:
                msg = _(
                    'mode "{0}" for "{1}" is not a valid mode, no modes are available'
                ).format(obj.mode, obj.name)
            raise DictConsistencyError(msg, 71, obj.xmlfiles)

    def _set_default_mode(
        self,
        family: "self.objectspace.family",
    ) -> None:
        children = self.objectspace.parents[family.path]
        if not children:
            return
        if self._has_mode(family):
            family_mode = family.mode
        else:
            family_mode = None
        leader = None
        for variable_path in children:
            variable = self.objectspace.paths[variable_path]
            if variable.type == "symlink":
                continue
            if leader is None and family.type == "leadership":
                leader = variable
                leader_mode = leader.mode
            if variable_path in self.objectspace.families:
                # set default mode a subfamily
                if family_mode and not self._has_mode(variable):
                    self._set_auto_mode(variable, family_mode)
            else:
                # set default mode to a variable
                self.valid_mode(variable)
                if leader:
                    self._set_default_mode_leader(leader, variable)
                self._set_default_mode_variable(variable, family_mode)
        if leader and leader_mode is not None:
            # here because follower can change leader mode
            self._set_auto_mode(family, leader_mode)

    def _has_mode(self, obj) -> bool:
        return obj.mode and not obj.path in self.mode_auto

    def _set_default_mode_variable(
        self,
        variable: "self.objectspace.variable",
        family_mode: Optional[str],
        check_level: bool = True,
    ) -> None:
        # auto_save variable is set to 'basic' mode
        # if its mode is not defined by the user
        if not self._has_mode(variable) and variable.auto_save is True:
            variable.mode = self.objectspace.modes_level[0]
        # mandatory variable without value is a basic variable
        elif (
            not self._has_mode(variable)
            and variable.mandatory is True
            and variable.default is None
            and variable.path not in self.objectspace.default_multi
        ):
            variable_mode = self.objectspace.modes_level[0]
            if (
                check_level
                and family_mode
                and self.modes[variable_mode] < self.modes[family_mode]
            ):
                msg = _(
                    'the variable "{0}" is mandatory so in "{1}" mode but family has the higher family mode "{2}"'
                ).format(variable.name, variable_mode, family_mode)
                raise DictConsistencyError(msg, 36, variable.xmlfiles)

            variable.mode = variable_mode
        elif family_mode and not self._has_mode(variable):
            self._set_auto_mode(variable, family_mode)

    def _set_auto_mode(
        self,
        obj,
        mode: str,
    ) -> None:
        obj.mode = mode
        self.mode_auto.append(obj.path)

    def _set_default_mode_leader(
        self,
        leader: "self.objectspace.variable",
        follower: "self.objectspace.variable",
    ) -> None:
        if leader == follower:
            # it's a leader
            if not leader.mode:
                self._set_auto_mode(leader, self.default_variable_mode)
            return
        if self._has_mode(follower):
            follower_mode = follower.mode
        else:
            follower_mode = self.default_variable_mode
        if self.modes[leader.mode] > self.modes[follower_mode]:
            if self._has_mode(follower) and not self._has_mode(leader):
                # if follower has mode but not the leader
                self._set_auto_mode(leader, follower_mode)
            else:
                # leader's mode is minimum level
                if self._has_mode(follower):
                    msg = _(
                        'the follower "{0}" is in "{1}" mode but leader have the higher mode "{2}"'
                    ).format(follower.name, follower_mode, leader.mode)
                    raise DictConsistencyError(msg, 63, follower.xmlfiles)
                self._set_auto_mode(follower, leader.mode)

    def _change_family_mode(
        self,
        family: "self.objectspace.family",
    ) -> None:
        if family.mode:
            family_mode = family.mode
        else:
            family_mode = self.default_family_mode
        min_variable_mode = self.objectspace.modes_level[-1]
        # change variable mode, but not if variables are not in a family
        is_leadership = family.type == "leadership"
        if family.path in self.objectspace.parents:
            for variable_path in self.objectspace.parents[family.path]:
                variable = self.objectspace.paths[variable_path]
                if variable.type == "symlink":
                    continue
                if variable_path in self.objectspace.families:
                    if not variable.mode:
                        variable.mode = self.default_family_mode
                else:
                    self._change_variable_mode(variable, family_mode, is_leadership)
                if self.modes[min_variable_mode] > self.modes[variable.mode]:
                    min_variable_mode = variable.mode
        if not family.mode:
            # set the lower variable mode to family
            self._set_auto_mode(family, min_variable_mode)
        if self.modes[family.mode] < self.modes[min_variable_mode]:
            msg = _(
                'the family "{0}" is in "{1}" mode but variables and families inside have the higher modes "{2}"'
            ).format(family.name, family.mode, min_variable_mode)
            raise DictConsistencyError(msg, 62, family.xmlfiles)

    def _change_variable_mode(
        self,
        variable,
        family_mode: str,
        is_follower: bool,
    ) -> None:
        if variable.mode:
            variable_mode = variable.mode
        else:
            variable_mode = self.default_variable_mode
        # none basic variable in high level family has to be in high level
        if not is_follower and self.modes[variable_mode] < self.modes[family_mode]:
            if self._has_mode(variable):
                msg = _(
                    'the variable "{0}" is in "{1}" mode but family has the higher family mode "{2}"'
                ).format(variable.name, variable_mode, family_mode)
                raise DictConsistencyError(msg, 61, variable.xmlfiles)
            self._set_auto_mode(variable, family_mode)
        if not variable.mode:
            variable.mode = variable_mode

    def convert_help(self):
        """Convert variable help"""
        for family in self.get_families():
            if not family.help:
                continue
            self.objectspace.informations.add(family.path, "help", family.help)
            del family.help



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['validators_general.int'] = "{% if _.int == int2 %}\nint and int2 must be different\n{% endif %}\n"
dict_env['validators_general.int_0'] = "{% if int3 is defined and _.int == int3 %}\nint and int3 must be different\n{% endif %}\n"
option_3 = IntOption(name="int2", doc="a second number", default=1, properties=frozenset({"mandatory", "standard"}), informations={'type': 'number'})
option_2 = IntOption(name="int", doc="a first number", validators=[Calculation(func['valid_with_jinja'], Params((), kwargs={'__internal_jinja': ParamValue("validators_general.int"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_7validators_variable_optional/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("validators"), '__internal_variable': ParamValue("general.int"), 'int2': ParamOption(option_3), '_.int': ParamSelfOption(whole=False)})), Calculation(func['valid_with_jinja'], Params((), kwargs={'__internal_jinja': ParamValue("validators_general.int_0"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_7validators_variable_optional/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("validators"), '__internal_variable': ParamValue("general.int"), '_.int': ParamSelfOption(whole=False)}))], properties=frozenset({"basic", "mandatory"}), informations={'type': 'number', 'test': (5,)})
optiondescription_1 = OptionDescription(name="general", doc="a family", children=[option_2, option_3], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['validators_rougail.var1'] = "{% if _.var1 == _.var2 %}\nvar1 must be different than var2\n{% endif %}\n"
option_3 = StrOption(name="var2", doc="A second variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_2 = StrOption(name="var1", doc="a first variable", default="oui", validators=[Calculation(func['valid_with_jinja'], Params((), kwargs={'__internal_jinja': ParamValue("validators_rougail.var1"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_5validators_differ/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("validators"), '__internal_variable': ParamValue("rougail.var1"), '_.var2': ParamOption(option_3, notraisepropertyerror=True), '_.var1': ParamSelfOption(whole=False)}))], properties=frozenset({"mandatory", "standard"}), informations={'type': 'string', 'test': ('another_value',)})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, option_3], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_1 = ChoiceOption(name="var1", doc="the first variable", values=("a", "b", "c"), properties=frozenset({"basic", "mandatory"}), informations={'type': 'choice'})
option_2 = ChoiceOption(name="var2", doc="the second variable", values=("a", "b", "c"), properties=frozenset({"basic", "mandatory"}), informations={'type': 'choice'})
option_3 = ChoiceOption(name="var3", doc="the third variable", values=("a", "b", "c", None), properties=frozenset({"standard"}), informations={'type': 'choice'})
option_4 = ChoiceOption(name="var4", doc="the forth variable", values=(None, "b", "c"), properties=frozenset({"standard"}), informations={'type': 'choice'})
option_5 = ChoiceOption(name="var5", doc="the fifth variable", values=("a", "b", "c"), default="a", properties=frozenset({"mandatory", "standard"}), informations={'type': 'choice'})
option_6 = ChoiceOption(name="var6", doc="the sixth variable", values=(1, 2, 3), default=1, properties=frozenset({"mandatory", "standard"}), informations={'type': 'choice'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1, option_2, option_3, option_4, option_5, option_6])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_1 = StrOption(name="condition", doc="condition", default="no", properties=frozenset({"mandatory", "normal"}))
optiondescription_8 = OptionDescription(name="rougail", doc="Rougail", children=[option_1], properties=frozenset({"normal"}))
option_3 = BoolOption(name="activate", doc="activate", default=Calculation(func.calc_value, Params((ParamValue(False)), kwargs={'default': ParamValue(True), 'condition_0': ParamOption(option_1, notraisepropertyerror=True), 'expected_0': ParamValue("yes")})))
option_4 = BoolOption(name="manage", doc="manage", default=False)
optiondescription_10 = OptionDescription(name="test_service", doc="test.service", children=[option_3, option_4])
optiondescription_10.impl_set_information('type', "none")
optiondescription_9 = OptionDescription(name="services", doc="services", children=[optiondescription_10], properties=frozenset({"hidden"}))
optiondescription_7 = OptionDescription(name="1", doc="1", children=[optiondescription_8, optiondescription_9])
option_2 = StrOption(name="condition", doc="condition", default="no", properties=frozenset({"mandatory", "normal"}))
optiondescription_12 = OptionDescription(name="rougail", doc="Rougail", children=[option_2], properties=frozenset({"normal"}))
option_5 = BoolOption(name="activate", doc="activate", default=Calculation(func.calc_value, Params((ParamValue(False)), kwargs={'default': ParamValue(True), 'condition_0': ParamOption(option_2, notraisepropertyerror=True), 'expected_0': ParamValue("yes")})))
option_6 = BoolOption(name="manage", doc="manage", default=False)
optiondescription_14 = OptionDescription(name="test_service", doc="test.service", children=[option_5, option_6])
optiondescription_14.impl_set_information('type', "none")
optiondescription_13 = OptionDescription(name="services", doc="services", children=[optiondescription_14], properties=frozenset({"hidden"}))
optiondescription_11 = OptionDescription(name="2", doc="2", children=[optiondescription_12, optiondescription_13])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_7, optiondescription_11])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries_old/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
from jinja2 import StrictUndefined, DictLoader
from jinja2.sandbox import SandboxedEnvironment
from rougail.annotator.variable import CONVERT_OPTION
def jinja_to_function(__internal_jinja, __internal_type, __internal_multi, **kwargs):
    kw = {}
    for key, value in kwargs.items():
        if '.' in key:
            c_kw = kw
            path, var = key.rsplit('.', 1)
            for subkey in path.split('.'):
                c_kw = c_kw.setdefault(subkey, {})
            c_kw[var] = value
        else:
            kw[key] = value
    values = ENV.get_template(__internal_jinja).render(kw)
    convert = CONVERT_OPTION[__internal_type].get('func', str)
    if __internal_multi:
        return [convert(val) for val in values.split(',')]
    return convert(values)
def valid_with_jinja(value, **kwargs):
    kwargs[kwargs.pop('__internal_key')] = value
    value = jinja_to_function(__internal_type='string', __internal_multi=False, **kwargs)
    if value:
        raise ValueError(value)
func.jinja_to_function = jinja_to_function
func.valid_with_jinja = valid_with_jinja
dict_env = {}
dict_env['rougail.general.leader.follower1'] = "{{ \"valfill\" | calc_val }}"
dict_env['rougail.general.leader.follower2'] = "{{ rougail.general.leader.follower1 | calc_val }}"
ENV = SandboxedEnvironment(loader=DictLoader(dict_env), undefined=StrictUndefined)
ENV.compile_templates('jinja_caches', zip=None)
option_3 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"mandatory", "normal"}))
option_5 = StrOption(name="leader", doc="leader", multi=True, properties=frozenset({"basic", "mandatory"}))
option_6 = StrOption(name="follower1", doc="follower1", multi=True, default=Calculation(func.jinja_to_function, Params((), kwargs={'__internal_jinja': ParamValue(rougail.general.leader.follower1), '__internal_type': ParamValue(string), '__internal_multi': ParamValue(True)})), properties=frozenset({"normal"}))
option_7 = StrOption(name="follower2", doc="follower2", multi=True, default=Calculation(func.jinja_to_function, Params((), kwargs={'__internal_jinja': ParamValue(rougail.general.leader.follower2), '__internal_type': ParamValue(string), '__internal_multi': ParamValue(True), 'rougail.general.leader.follower1': ParamOption(option_6)})), properties=frozenset({"normal"}))
optiondescription_4 = Leadership(name="leader", doc="leader", children=[option_5, option_6, option_7], properties=frozenset({"basic"}))
optiondescription_2 = OptionDescription(name="general", doc="general", children=[option_3, optiondescription_4], properties=frozenset({"basic"}))
optiondescription_1 = OptionDescription(name="rougail", doc="rougail", children=[optiondescription_2], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['disabled_rougail.{{ identifier }}_dyn.var4'] = "{% if 'val4' not in __.var %}\nval4 is not a valid value\n{% endif %}\n"
option_2 = StrOption(name="var", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string', 'test': ('val1', 'val2', 'val3', 'val4')})
option_4 = StrOption(name="var1", doc="a variable 1", default=Calculation(func['calc_value'], Params((ParamIdentifier()))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_5 = StrOption(name="var2", doc="a variable 2", default=Calculation(func['calc_value'], Params((ParamOption(option_4)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_6 = StrOption(name="var3", doc="a variable 3", default=Calculation(func['calc_value'], Params((ParamOption(option_4)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_7 = StrOption(name="var4", doc="a variable 4", default=Calculation(func['calc_value'], Params((ParamDynOption(option_4, ["val4"])))), properties=frozenset({"mandatory", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("disabled")), kwargs={'__internal_jinja': ParamValue("disabled_rougail.{{ identifier }}_dyn.var4"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_unknown_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("disabled"), '__internal_variable': ParamValue("rougail.{{ identifier }}_dyn.var4"), 'when': ParamValue(True), 'inverse': ParamValue(False), '__.var': ParamOption(option_2, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_3 = ConvertDynOptionDescription(name="{{ identifier }}_dyn", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_2)))), children=[option_4, option_5, option_6, option_7], properties=frozenset({"standard"}), informations={'dynamic_variable': 'rougail.var'})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, optiondescription_3], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_1 = StrOption(name="my_calculated_variable", doc="my_calculated_variable", multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_1 = StrOption(name="condition", doc="condition", default="no", properties=frozenset({"mandatory", "normal"}))
option_2 = IPOption(name="nut_monitor_host", doc="nut_monitor_host", default="192.168.0.1", allow_reserved=True, properties=frozenset({"mandatory", "normal"}))
optiondescription_18 = OptionDescription(name="rougail", doc="Rougail", children=[option_1, option_2], properties=frozenset({"normal"}))
option_9 = SymLinkOption(name="name", opt=option_2)
option_8 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_7 = OptionDescription(name="nut_monitor_host", doc="nut_monitor_host", children=[option_9, option_8])
optiondescription_6 = OptionDescription(name="ip", doc="ip", children=[optiondescription_7])
option_5 = BoolOption(name="activate", doc="activate", default=Calculation(func.calc_value, Params((ParamValue(False)), kwargs={'default': ParamValue(True), 'condition_0': ParamOption(option_1, notraisepropertyerror=True), 'expected_0': ParamValue("no")})))
option_10 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_20 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_6, option_5, option_10])
optiondescription_20.impl_set_information('type', "service")
optiondescription_19 = OptionDescription(name="services", doc="services", children=[optiondescription_20], properties=frozenset({"hidden"}))
optiondescription_17 = OptionDescription(name="1", doc="1", children=[optiondescription_18, optiondescription_19])
option_3 = StrOption(name="condition", doc="condition", default="no", properties=frozenset({"mandatory", "normal"}))
option_4 = IPOption(name="nut_monitor_host", doc="nut_monitor_host", default="192.168.0.1", allow_reserved=True, properties=frozenset({"mandatory", "normal"}))
optiondescription_22 = OptionDescription(name="rougail", doc="Rougail", children=[option_3, option_4], properties=frozenset({"normal"}))
option_15 = SymLinkOption(name="name", opt=option_4)
option_14 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_13 = OptionDescription(name="nut_monitor_host", doc="nut_monitor_host", children=[option_15, option_14])
optiondescription_12 = OptionDescription(name="ip", doc="ip", children=[optiondescription_13])
option_11 = BoolOption(name="activate", doc="activate", default=Calculation(func.calc_value, Params((ParamValue(False)), kwargs={'default': ParamValue(True), 'condition_0': ParamOption(option_3, notraisepropertyerror=True), 'expected_0': ParamValue("no")})))
option_16 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_24 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_12, option_11, option_16])
optiondescription_24.impl_set_information('type', "service")
optiondescription_23 = OptionDescription(name="services", doc="services", children=[optiondescription_24], properties=frozenset({"hidden"}))
optiondescription_21 = OptionDescription(name="2", doc="2", children=[optiondescription_22, optiondescription_23])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_17, optiondescription_21])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_1 = StrOption(name="variable", doc="a variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_3 = StrOption(name="variable1", doc="a first variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_5 = StrOption(name="variable", doc="a variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_4 = OptionDescription(name="subfamily", doc="a sub family", children=[option_5], properties=frozenset({"basic"}))
option_6 = StrOption(name="variable2", doc="a second variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="family", doc="a family", children=[option_3, optiondescription_4, option_6], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1, optiondescription_2])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_leader.follower2'] = "{{ _.follower1 }}\n"
option_2 = StrOption(name="leader", doc="a leader", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_3 = StrOption(name="follower1", doc="a follower", multi=True, default_multi="value", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_4 = StrOption(name="follower2", doc="a second follower", multi=True, default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_leader.follower2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/40_0leadership_follower_default_calculation/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("leader.follower2"), '_.follower1': ParamOption(option_3, notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_1 = Leadership(name="leader", doc="a leadership", children=[option_2, option_3, option_4], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['choices_1.rougail.var'] = "{% for item in trange(0, 10) %}\n{{ item }}\n{%- endfor %}\n"
dict_env['choices_2.rougail.var'] = "{% for item in trange(0, 10) %}\n{{ item }}\n{%- endfor %}\n"
option_3 = ChoiceOption(name="var", doc="a variable", values=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("choices_1.rougail.var"), '__internal_type': ParamValue("number"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/00_9default_integer/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("choices"), '__internal_variable': ParamValue("1.rougail.var")})), default=9, properties=frozenset({"mandatory", "standard"}), informations={'type': 'choice'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_6 = ChoiceOption(name="var", doc="a variable", values=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("choices_2.rougail.var"), '__internal_type': ParamValue("number"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/00_9default_integer/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("choices"), '__internal_variable': ParamValue("2.rougail.var")})), default=9, properties=frozenset({"mandatory", "standard"}), informations={'type': 'choice'})
optiondescription_5 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_6], properties=frozenset({"standard"}))
optiondescription_4 = OptionDescription(name="2", doc="2", children=[optiondescription_5], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_4])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_3 = NetmaskOption(name="nut_monitor_netmask", doc="nut_monitor_netmask", multi=True, default=['255.255.255.0'], properties=frozenset({"mandatory", "normal"}))
option_4 = NetworkOption(name="nut_monitor_host", doc="nut_monitor_host", multi=True, default_multi="192.168.1.0", properties=frozenset({"mandatory", "normal"}))
optiondescription_2 = Leadership(name="nut_monitor_netmask", doc="nut_monitor_netmask", children=[option_3, option_4], properties=frozenset({"normal"}))
optiondescription_1 = OptionDescription(name="general", doc="général", children=[optiondescription_2], properties=frozenset({"normal"}))
optiondescription_24 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"normal"}))
option_13 = SymLinkOption(name="name", opt=option_4)
option_14 = SymLinkOption(name="netmask", opt=option_3)
option_12 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_11 = OptionDescription(name="nut_monitor_host", doc="nut_monitor_host", children=[option_13, option_14, option_12])
optiondescription_10 = OptionDescription(name="ip", doc="ip", children=[optiondescription_11])
option_9 = BoolOption(name="activate", doc="activate", default=True)
option_15 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_26 = OptionDescription(name="ntp_service", doc="ntp.service", children=[optiondescription_10, option_9, option_15])
optiondescription_26.impl_set_information('type', "service")
optiondescription_25 = OptionDescription(name="services", doc="services", children=[optiondescription_26], properties=frozenset({"hidden"}))
optiondescription_23 = OptionDescription(name="1", doc="1", children=[optiondescription_24, optiondescription_25])
option_7 = NetmaskOption(name="nut_monitor_netmask", doc="nut_monitor_netmask", multi=True, default=['255.255.255.0'], properties=frozenset({"mandatory", "normal"}))
option_8 = NetworkOption(name="nut_monitor_host", doc="nut_monitor_host", multi=True, default_multi="192.168.1.0", properties=frozenset({"mandatory", "normal"}))
optiondescription_6 = Leadership(name="nut_monitor_netmask", doc="nut_monitor_netmask", children=[option_7, option_8], properties=frozenset({"normal"}))
optiondescription_5 = OptionDescription(name="general", doc="général", children=[optiondescription_6], properties=frozenset({"normal"}))
optiondescription_28 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_5], properties=frozenset({"normal"}))
option_20 = SymLinkOption(name="name", opt=option_8)
option_21 = SymLinkOption(name="netmask", opt=option_7)
option_19 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_18 = OptionDescription(name="nut_monitor_host", doc="nut_monitor_host", children=[option_20, option_21, option_19])
optiondescription_17 = OptionDescription(name="ip", doc="ip", children=[optiondescription_18])
option_16 = BoolOption(name="activate", doc="activate", default=True)
option_22 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_30 = OptionDescription(name="ntp_service", doc="ntp.service", children=[optiondescription_17, option_16, option_22])
optiondescription_30.impl_set_information('type', "service")
optiondescription_29 = OptionDescription(name="services", doc="services", children=[optiondescription_30], properties=frozenset({"hidden"}))
optiondescription_27 = OptionDescription(name="2", doc="2", children=[optiondescription_28, optiondescription_29])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_23, optiondescription_27])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['validators_var3'] = "{% if _.var3 == _.var2 %}\nvar3 must be different than var2\n{% endif %}\n"
option_1 = StrOption(name="var1", doc="a first variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_2 = StrOption(name="var2", doc="a second variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_3 = StrOption(name="var3", doc="a third variable", default="yes", validators=[Calculation(func['valid_with_jinja'], Params((), kwargs={'__internal_jinja': ParamValue("validators_var3"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/20_0validators_differ_redefine/dictionaries/rougail/01-base.yml']), '__internal_attribute': ParamValue("validators"), '__internal_variable': ParamValue("var3"), '_.var3': ParamSelfOption(whole=False), '_.var2': ParamOption(option_2, notraisepropertyerror=True)}))], properties=frozenset({"mandatory", "standard"}), informations={'type': 'string', 'test': ('yes',)})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1, option_2, option_3])
